export function reduce (expr) {
	if (!is_application(expr)) {
		return expr;
	}

	const result = apply(reduce(expr[0]), reduce(expr[1]));

	// eager evaluation:
	// TODO: use trampolining to avoid stack overflow.
	if (is_application(result)) {
		return reduce(result);
	}

	return result;
}

function is_application (expr) {
	return expr.length === 2
		&& !is_function(expr);
}

function is_function (expr) {
	return expr[0] === 0;
}

function apply (fn, arg) {
	const body = fn[1];
	return replace(body, arg);
}

function replace (expr, val, depth = 1) {
	const next_depth = is_function(expr) ? depth + 1 : depth;

	return Array.isArray(expr)
		? expr.map((item) => replace(item, val, next_depth))
	: expr === depth
		? val
	: expr;
}
