import Line from "./Line.mjs";
import Expr from "./Expr.mjs";

import { parse_line } from "./parse.mjs";
import { compile_line } from "./compile.mjs";
import { transpile } from "./transpile.mjs";
import { reduce } from "./reduce.mjs";

import readline from "node:readline";

const rl = readline.createInterface({
	input: process.stdin,
	output: process.stdout,
});



let running = true;

let env = new Map();

function print (str) {
	console.log(str);
}



const specials = new Map();

specials.set("!.exit", _ => running = false);

specials.set("!.env", _ => {
	const lines = [...env.entries()].map(([name, fn]) => name + " = " + prettify(fn));
	print(lines.join("\n"));
});

specials.set("!.clear", _ => console.clear());

specials.set("!.load combinators", _ => {
	combinators.forEach(c => run(c, false));
});



function run (line, output = true) {
	if (specials.has(line)) {
		specials.get(line)();
		return;
	}

	const parsed = parse_line(line);
	const compiled = compile_line(env, parsed);

	if (!compiled.valid) {
		if (output) print("invalid line!");
		return;
	}

	env = compiled.env;

	if (!output) return;

	if (parsed.type === Line.Expression) {
		print(prettify(reduce(compiled.value)));

		// js function that's equivalent to the entered lambda function.
		// const fn = transpile(compiled.value));
	} else if (parsed.type === Line.Binding) {
		print(parsed.name + " = " + prettify(compiled.fn));
	}
}

function evaluate (fn) {
	return transpile(fn[0])(fn[1]);
}

function prettify (code) {
	if (typeof code === "number") {
		return code.toString();
	} else if (code[0] === 0) {
		return "(λ" + prettify(code[1]) + ")";
	} else {
		return "(" + prettify(code[0]) + " " + prettify(code[1]) + ")";
	}
}

function read (callback) {
	rl.question("> ", callback);
}

function repl () {
	read(line => {
		run(line);

		if (running) {
			setImmediate(repl);
		} else {
			rl.close();
		}
	});
}

print("welcome to catgirl calculus!");
print("uwu");
repl();
